var utils = require("../utils");

/**
 * Enrich generated data from signatures of source files (if found).
 * Currently enriching only method meta object and hasshes if method is async.
 * @param generatedSignature
 * @param sourceSignatures
 * @param logger
 * @constructor
 */

var CODE_ELEMENT_TYPES = {
    BRANCH: "branch",
    METHOD: "method"
};
function SourceDataEnricher(generatedSignature, sourceSignatures, logger) {
    this.generatedSignature = generatedSignature;
    this.sourceSignatures = sourceSignatures;
    this.logger = logger;
}

/**
 * Iterate over generated and source methods, find match by source position and enrich method data.
 */
SourceDataEnricher.prototype.enrichData = function(){
    var context = this;
    this.sourceSignatures.forEach(function (sourceSig) {
        if(!sourceSig) {
            context.logger.error("Source signature is undefined, skipping enrichment");
            return;
        }
        var sourcePositionMap = context.generatedSignature.sourcePositionMappers[sourceSig.absolutePath];
        context.enrichMethods(sourcePositionMap, sourceSig);
        context.enrichBranches(sourcePositionMap, sourceSig);
    })
}

SourceDataEnricher.prototype.enrichMethods = function(sourcePositionMap, sourceSig) {
    var context = this;
    var generatedMethodsArray = context.generatedSignature.methods;
    var sourceMethodsArray = sourceSig.methods;
    this.enrichProcedure(sourcePositionMap.getMethodsPosition(), sourceSig, generatedMethodsArray, sourceMethodsArray, sourcePositionMap,
        sourceSig.positionToMethodMap,CODE_ELEMENT_TYPES.METHOD);
}

SourceDataEnricher.prototype.enrichBranches = function(sourcePositionMap, sourceSig) {
    var context = this;
    var generatedBranchArray = context.generatedSignature.branches;
    var sourceBranchesArray = sourceSig.branches;
    this.enrichProcedure(sourcePositionMap.getBranchesPosition(), sourceSig, generatedBranchArray, sourceBranchesArray, sourcePositionMap,
        sourceSig.positionToBranchMap, CODE_ELEMENT_TYPES.BRANCH);
}

SourceDataEnricher.prototype.handleAutoGeneratedElement = function (elementType, position, sourceSig, generatedElement) {
    this.logger.info("Couldn't find %s at position '%s' in file '%s', but it got generate as %s." +
        "marked as auto generate code", elementType, position, sourceSig.absolutePath, elementType);
    generatedElement.techSpecificInfo["autoGenerated"] = true;
}

SourceDataEnricher.prototype.enrichProcedure= function(srcPositionsArr, sourceSig, genElementsArr, srcElementsArr,
                                                        srcPositionToGenElementIdx, srcPositionToSrcElementIdx, codeElementType) {
    var context = this;
    srcPositionsArr.forEach(function (srcPosition) {
        var sourceElementIndex = srcPositionToSrcElementIdx[srcPosition];
        var getGenIdxMethod = codeElementType ===  CODE_ELEMENT_TYPES.METHOD ? srcPositionToGenElementIdx.getMethodIndex : srcPositionToGenElementIdx.getBranchIndex;
        var generatedElementIndex = getGenIdxMethod.call(srcPositionToGenElementIdx, srcPosition);
        var generatedElement = genElementsArr[generatedElementIndex];
          var sourceElement = srcElementsArr[sourceElementIndex];
        if (generatedElement == null) {
            context.logger.warn("%s with source position '%s' in file '%s', is null", codeElementType, srcPosition, context.generatedSignature.absolutePath);
        }
        else if (sourceElement == null) {
            context.handleAutoGeneratedElement(codeElementType, srcPosition, sourceSig, generatedElement);
        }
        else if(codeElementType == CODE_ELEMENT_TYPES.METHOD) {
            updateMethodInfo(generatedElement, sourceElement);
        }
    })
}

function updateMethodInfo(generatedMethod, sourceMethod) {
    generatedMethod.meta = utils.cloneObject(sourceMethod.meta);
    generatedMethod.techSpecificInfo = utils.cloneObject(sourceMethod.techSpecificInfo);
    if (sourceMethod.techSpecificInfo.isAsync) {
        generatedMethod.srcData.hash = sourceMethod.generatedData.hash;
        generatedMethod.srcData.sigHash = sourceMethod.generatedData.sigHash;
    }
}
module.exports = SourceDataEnricher;